package raytracer.graphics;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.function.Consumer;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL20;
import org.lwjgl.util.vector.Matrix4f;

import raytracer.RayTracer;

public class Shader {
	
	private static final ArrayList<Shader> shaders = new ArrayList<Shader>();
	public static Shader rayShader;
	public static boolean lastCompileFailed = false;
	
	public static void init() {
		ShaderBuilder rayShaderBuilder = ShaderBuilder.start("rayShader").load("frag");
		if (!rayShaderBuilder.hasError()) {
			cleanUp();
			rayShader = rayShaderBuilder.attatchAll().linkAndFinish();
			lastCompileFailed = false;
		} else lastCompileFailed = true;
	}
	
	private static FloatBuffer matrixBuffer = BufferUtils.createFloatBuffer(16);
	
	private HashMap<String, Integer> uniformLocations = new HashMap<String, Integer>();
	private HashMap<String, Integer> attributeLocations = new HashMap<String, Integer>();
	
	public final int shaderID;
	
	Shader(int id) {
		shaderID = id;
		shaders.add(this);
	}
	
	public Shader(String name, String... attributeBinds) {
		this("res/shaders/" + name + ".vert", "res/shaders/" + name + ".frag", attributeBinds);
	}
	
	public Shader(String vertPath, String fragPath, String... attributeBinds) {
		shaderID = GL20.glCreateProgram();
		load(vertPath, fragPath, attributeBinds);
		shaders.add(this);
	}
	
	public int getUniformLocation(String name) {
		Integer loc = uniformLocations.get(name);
		if (loc == null) {
			loc = GL20.glGetUniformLocation(shaderID, name);
			uniformLocations.put(name, loc);
		}
		return loc;
	}
	
	public int getAttribLocation(String name) {
		Integer loc = attributeLocations.get(name);
		if (loc == null) {
			loc = GL20.glGetAttribLocation(shaderID, name);
			attributeLocations.put(name, loc);
		}
		return loc;
	}
	
	public void bindAttribute(int loc, String name) {
		GL20.glBindAttribLocation(shaderID, loc, name);
	}
	
	public void setUniform1i(String name, int a) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform1i(loc, a);
	}
	
	public void setUniform1f(String name, float a) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform1f(loc, a);
	}
	
	public void setUniform2f(String name, float a, float b) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform2f(loc, a, b);
	}
	
	public void setUniform3f(String name, float a, float b, float c) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform3f(loc, a, b, c);
	}
	
	public void setUniform4f(String name, float a, float b, float c, float d) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform4f(loc, a, b, c, d);
	}
	
	public void setUniformMat4(String name, Matrix4f matrix) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		matrix.store(matrixBuffer);
		matrixBuffer.flip();
		GL20.glUniformMatrix4(loc, false, matrixBuffer);
		matrixBuffer.clear(); // maybe unnecessary
	}
	
	public void enable() {
		GL20.glUseProgram(shaderID);
	}
	
	public void delete() {
		GL20.glDeleteProgram(shaderID);
	}
	
	public static void forEachShader(Consumer<Shader> action) {
		shaders.forEach(s -> {
			s.enable();
			action.accept(s);
		});
	}
	
	private void load(String vertPath, String fragPath, String... attributeBinds) {
		int vertexShader = createShader(GL20.GL_VERTEX_SHADER, vertPath);
		int fragmentShader = createShader(GL20.GL_FRAGMENT_SHADER, fragPath);
		
		GL20.glAttachShader(shaderID, vertexShader);
		GL20.glAttachShader(shaderID, fragmentShader);
		
		int index = 0;
		for (String s : attributeBinds)
			bindAttribute(index++, s);
		
		GL20.glLinkProgram(shaderID);
		
		if (GL20.glGetProgrami(shaderID, GL20.GL_LINK_STATUS) == GL11.GL_FALSE) {
			System.out.println("Link error: " + GL20.glGetProgramInfoLog(shaderID, 1024));
		}
		
		GL20.glValidateProgram(shaderID);
		
		GL20.glDetachShader(shaderID, vertexShader);
		GL20.glDetachShader(shaderID, fragmentShader);
		
		GL20.glDeleteShader(vertexShader);
		GL20.glDeleteShader(fragmentShader);
	}
	
	private int createShader(int type, String path) {
		int shaderID = GL20.glCreateShader(type);
		StringBuilder source = new StringBuilder();
		
		try {
			BufferedReader reader = new BufferedReader(new FileReader(path));
			String line;
			while ((line = reader.readLine()) != null)
				source.append(line).append('\n');
			reader.close();
		} catch (IOException e) {
			e.printStackTrace();
			RayTracer.destroy();
			System.exit(1);
		}
		
		GL20.glShaderSource(shaderID, source);
		GL20.glCompileShader(shaderID);
		if (GL20.glGetShaderi(shaderID, GL20.GL_COMPILE_STATUS) == GL11.GL_FALSE) System.err.println((type == GL20.GL_VERTEX_SHADER ? "Vertexshader" : type == GL20.GL_FRAGMENT_SHADER ? "Fragmentshader" : "Shader") + " failed to compile.\nError log:\n" + GL20.glGetShaderInfoLog(shaderID, GL20.glGetShaderi(shaderID, GL20.GL_INFO_LOG_LENGTH)));
		return shaderID;
	}
	
	public static void cleanUp() {
		GL20.glUseProgram(0);
		for (Shader s : shaders)
			s.delete();
		shaders.clear();
	}
	
}
